/*
 * symplectic_kernel.h
 *
 * Header files for symplectic_basis, all .c files include this header file.
 */


#ifndef SYMPLECTIC_KERNEL_H
#define SYMPLECTIC_KERNEL_H

#include <stdio.h>
#include <string.h>

#include "kernel.h"

#include "symplectic_prototypes.h"
#include "symplectic_typedefs.h"

#endif /* SYMPLECTIC_KERNEL_H */
